// this file contains: interpreter pluggable primitive support functions
//
// author: Colin Sarsfield
//
// Please read the file named License for license information.
// 

#include <image.h>
#include <Path.h>

extern "C" {
	#include "sq.h"
}

#include "sqBeApp.h"
#include "sqBeExternalPrims.h"

//image_id squeakVMLib = B_ERROR;  /* connection to the VM itself as a shared library */

/*typedef struct LoadedModule 
	{
	char moduleName[255];
	image_id id;
	struct LoadedModule *next;
	} LoadedModule;
	
LoadedModule *FirstLoadedModule = NULL;

void FreeLoadedModules()
	{
	LoadedModule *delModule, *currentModule = FirstLoadedModule;
	
	while (currentModule)
		{
		delModule = currentModule;
		currentModule = currentModule->next;
		unload_add_on(delModule->id);
		delete delModule;
		}
	FirstLoadedModule = NULL;
	}*/

/*** Function Type Declaration ***/
//typedef int (*RecordVMProxyProc)(struct VirtualMachine *interpreterProxy);

/* The next three functions must be implemented by sqXYZExternalPrims.c */
/* ioLoadModule:
	Load a module from disk.
	WARNING: this always loads a *new* module. Don't even attempt to find a loaded one.
	WARNING: never primitiveFail() within, just return 0
*/
int ioLoadModule(char *pluginName) 
{
	char modulePath[B_PATH_NAME_LENGTH];
	image_id id;
	strcpy(modulePath, VMPath);
	strcat(modulePath, pluginName);
	id = load_add_on(modulePath);
	if (id <= 0) {
		//printf("couldn't load module %s\n", pluginName);
		return 0;
	} else {
		//printf("loaded module %s (%d)\n", pluginName, id);
		return id;
	}
}

/* ioFindExternalFunctionIn:
	Find the function with the given name in the moduleHandle.
	WARNING: never primitiveFail() within, just return 0.
*/
int ioFindExternalFunctionIn(char *lookupName, int moduleHandle) 
{
	void *functionPtr;
	if (get_image_symbol((image_id)moduleHandle, lookupName, 
			B_SYMBOL_TYPE_TEXT, &functionPtr) == B_OK) {
		//printf("Found symbol %s\n", lookupName);
		return (int) functionPtr;
	} else {
		//printf("Didn't find symbol %s\n", lookupName);
		return 0;
	}
}

/* ioFreeModule:
	Free the module with the associated handle.
	WARNING: never primitiveFail() within, just return 0.
*/
int ioFreeModule(int moduleHandle) 
{
	unload_add_on(moduleHandle);
	//printf("unloaded module %d\n", moduleHandle);
	return 1;
}
/*int ioLoadExternalFunctionOfLengthFromModuleOfLength(
  int functionNameIndex, int functionNameLength,
  int moduleNameIndex, int moduleNameLength) { return 0; }
int ioUnloadModuleOfLength(int moduleNameIndex, int moduleNameLength) { return 0; }
int ioLoadFunctionFrom(char *functionName, char *pluginName) { return 0; }
int ioShutdownAllModules(void) { return 0; }
int ioUnloadModule(char *) { return 0; }
char *ioListBuiltinModule(int moduleIndex) { return 0; }
char *ioListLoadedModule(int moduleIndex) { return 0; }
int ioLoadModuleOfLength(int moduleNameIndex, int moduleNameLength){ return 0; }
int ioLoadSymbolOfLengthFromModule(int functionNameIndex, int functionNameLength, int moduleHandle){ return 0; }
*/
/*
int ioLoadExternalFunctionOfLengthFromModuleOfLength(
  int functionNameIndex, int functionNameLength, int moduleNameIndex, int moduleNameLength) {
	// this could be cleaned up & speeded up by creating a cache of
	// image_id's by moduleName
	char functionName[255], moduleName[255];
	void *functionPtr;
	LoadedModule *currentModule = FirstLoadedModule;
	image_id id = B_ERROR;
	int i, x;

	for (i = 0; i < functionNameLength; i++) {
		functionName[i] = ((char *) functionNameIndex)[i];
	}
	functionName[functionNameLength] = 0;
	
	if (moduleNameLength > 0) {
		strcpy(moduleName, VMPath);
		x = strlen(moduleName);
		for (i = 0; i < moduleNameLength; i++) {
			moduleName[i + x] = ((char *) moduleNameIndex)[i];
		}
		moduleName[moduleNameLength + x] = 0;
		//printf("'%s'>>'%s'\n", moduleName, functionName);
	} else {
		*moduleName = '\0';
		//printf("VM>>'%s'\n", functionName);
	}
	

	if (moduleNameLength > 0) 
		{
		// is module already loaded?
		while (currentModule)
			{
			if (!strcmp(currentModule->moduleName, moduleName))
				{
				id = currentModule->id;
				//printf("Found already loaded module\n");
				break;
				}
			currentModule = currentModule->next;
			}
		// if not already loaded, then load it!
		if (!currentModule)
			{
			id = load_add_on(moduleName);
			// if loaded ok, then add it to the list
			if (id > 0)
				{
				//printf("Successfully loaded module %s\n", moduleName);
				currentModule = new LoadedModule;
				currentModule->id = id;
				strcpy(currentModule->moduleName, moduleName);
				currentModule->next = FirstLoadedModule;
				FirstLoadedModule = currentModule;
				}						
			}
		} 
	else {
		if (squeakVMLib == B_ERROR) {
			image_info info; 
			int32 cookie = 0; 
   			while (get_next_image_info(0, &cookie, &info) == B_OK)
   				if (info.type == B_APP_IMAGE)
   					squeakVMLib = info.id;
		}
		id = squeakVMLib;
	}

	#if false
		// dump symbols
		int32 n = 1;
		const int32 nl = 256;
		char name[nl];
		int32 name_length = nl;
		int32 symbol_type;
		void *location;
		status_t status = B_OK;
		do {
			name_length = nl;
			status = get_nth_image_symbol(id, n, name, &name_length, &symbol_type, &location);
			if (status != B_OK) break;
			n++;
			printf("symbol: %s\n", name);
		} while (true);
	#endif
	
	if (id == B_ERROR) return success(false);  

	if (id != squeakVMLib) {
		if (get_image_symbol(id, "setInterpreter", B_SYMBOL_TYPE_TEXT, &functionPtr) == B_OK) {
			//printf("Found setInterpreter\n");
			((RecordVMProxyProc) functionPtr)(sqGetInterpreterProxy());
		} else {
			//printf("Didn't find setInterpreter\n");
			return success(false);
		}
	}

	if (get_image_symbol(id, functionName, B_SYMBOL_TYPE_TEXT, &functionPtr) == B_OK) {
		//printf("Found symbol\n");
		return (int) functionPtr;
	} else {
		//printf("Didn't find symbol\n");
		return success(false);
	}
}
*/
